/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "dix.h"
#include "osdep.h"
#include "dixstruct.h"
#include "scrnintstr.h"
#include "windowstr.h"

#include "Agent.h"
#include "Handlers.h"
#include "Display.h"
#include "Events.h"
#include "Client.h"
#include "Reconnect.h"
#include "Options.h"
#include "Dialog.h"
#include "Drawable.h"
#include "Pixmap.h"
#include "Splash.h"

#include "NXlib.h"

/*
 * Set here the required log level.
 */

#define PANIC
#undef  WARNING
#undef  TEST
#undef  DEBUG
#undef  DUMP

/*
 * If not defined, flush immediately
 * upon entering the block handler.
 */

#define FLUSH_AFTER_MULTIPLE_READS

int nxagentFlushCount   = 0;
int nxagentFlushPending = 0;

/*
 * Introduce a small delay after each
 * loop if the session is down. The
 * value is in milliseconds.
 */

#define LOOP_DELAY_IF_DOWN       50

/*
 * Break out of the select if there is
 * any drawable to synchronize. Value
 * is in milliseconds.
 */

#define SYNCHRONIZATION_TIMEOUT  2000

/*
 * If defined, check if we need a null
 * timeout after a client wakeup.
 */

#undef CHECK_RESTARTED_CLIENTS

#ifdef FLUSH_AFTER_MULTIPLE_READS

static int nxagentFlushCoalescence = 0 * 1000;
static int nxagentFlushCoalescenceLimit = 4;

#endif

int nxagentXdmcpUp = 0;
int nxagentXdmcpAlertUp = 0;

/*
 * Timestamp taken the last time we ran
 * the drawable synchronization loop.
 */

unsigned long nxagentLastSynchronization = 0;

/*
 * Called just before blocking waiting
 * for more data from the network.
 */

void nxagentBlockHandler(pointer data, struct timeval **timeout, pointer mask)
{
  static struct timeval zero;
  static unsigned long  now;

  #ifdef CHECK_RESTARTED_CLIENTS

  int i;

  #endif

  now = GetTimeInMillis();
  
  if (nxagentNeedConnectionChange())
  {
    nxagentHandleConnectionChanges();
  }

  /*
   * We need this here because some window
   * configuration changes can be generated
   * by the X server outside the control of
   * the DIX.
   */

  nxagentFlushConfigureWindow();

  if (nxagentOption(Rootless) &&
          nxagentLastWindowDestroyed && nxagentRootlessDialogPid == 0 &&
              now > nxagentLastWindowDestroyedTime + 30 * 1000)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentBlockHandler: No application running. Closing the session.\n");
    #endif

    nxagentSetGiveUp();
  }

  #ifdef TEST
  if (nxagentLastWindowDestroyed == 1)
  {
    fprintf(stderr, "nxagentBlockHandler: Elapsed time [%u].\n",
                now - nxagentLastWindowDestroyedTime);
  }
  #endif

  /*
   * Dispatch to the clients the events that may
   * have become available.
   */

  if (XEventsQueued(nxagentDisplay, QueuedAlready))
  {
    #ifdef TEST
    fprintf(stderr, "nxagentBlockHandler: Reading X events queued.\n");
    #endif

    nxagentDispatchEvents(NULL);
  }

  /*
   * Slow down the agent if the session is
   * not connected to a valid display.
   */

  if (NXQueryDisplayError(nxagentDisplay) == 1)
  {
    usleep(LOOP_DELAY_IF_DOWN * 1000);

    /*
     * Avoid waiting for more input if the
     * display is down.
     */

    nxagentFlush = 1;
  }

  /*
   * Update the shadow display. This is
   * only for test purposes.
   */

  #ifdef DUMP

  nxagentPixmapOnShadowDisplay(NULL);

  nxagentFbOnShadowDisplay();

  #endif

  /*
   * If any of the restarted clients had requests
   * in input we'll need to enter the select with
   * a null timeout, or we will block until any
   * other client becomes available.
   */

  #ifdef CHECK_RESTARTED_CLIENTS

  for (i = 1; i < currentMaxClients; i++)
  {
    if (clients[i] != NULL && clients[i] -> osPrivate != NULL &&
           nxagentNeedWakeup(clients[i]) == 0)
    {
      int fd = ((OsCommPtr) clients[i] -> osPrivate) -> fd;

      if (FD_ISSET(fd, &ClientsWithInput))
      {
        #ifdef WARNING
        fprintf(stderr, "nxagentBlockHandler: WARNING! Client [%d] with fd [%d] has input.\n",
                    clients[i] -> index, fd);
        #endif

        #ifdef DEBUG
        fprintf(stderr, "nxagentBlockHandler: Setting a null timeout with former timeout [%ld] Ms.\n",
                    (*timeout) -> tv_sec * 1000 + (*timeout) -> tv_usec / 1000);
        #endif

        if (*timeout != NULL)
        {
          (*timeout) -> tv_sec  = 0;
          (*timeout) -> tv_usec = 0;
        }
        else
        {
          zero.tv_sec  = 0;
          zero.tv_usec = 0;

          *timeout = &zero;
        }
      }
    }
  }

  #endif /* #ifdef CHECK_RESTARTED_CLIENTS */

  #ifdef FLUSH_AFTER_MULTIPLE_READS

  /*
   * Write any remaining data to the NX transport.
   * If we didn't produce enough data for a flush
   * let's try to read some more from any client
   * that may have became readable.
   */

  NXFlushDisplay(nxagentDisplay, NXFlushBuffer);

  if (nxagentFlush == 1)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentBlockHandler: Flushing with [%d] bytes count [%d] on demand "
                "of the NX transport.\n", NXQueryDisplay(nxagentDisplay, NXDisplayFlushable),
                    nxagentFlushCount);
    #endif

    nxagentFlush = 0;

    nxagentFlushCount = 0;

    nxagentFlushPending = 0;

    NXFlushDisplay(nxagentDisplay, NXFlushLink);
  }
  else
  {
    int pending = NXQueryDisplay(nxagentDisplay, NXDisplayFlushable);

    if (pending == 0)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentBlockHandler: Resetting the counter with no data available.\n");
      #endif

      nxagentFlushCount = 0;

      nxagentFlushPending = 0;
    }
    else if (nxagentFlushPending > 0 &&
                 nxagentFlushPending == pending)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentBlockHandler: Flushing with same [%d] bytes and count [%d].\n",
                  nxagentFlushPending, nxagentFlushCount);
      #endif

      nxagentFlush = 0;

      nxagentFlushCount = 0;

      nxagentFlushPending = 0;

      NXFlushDisplay(nxagentDisplay, NXFlushLink);
    }
    else
    {
      nxagentFlushPending = pending;

      nxagentFlushCount++;

      if (nxagentFlushCount > nxagentFlushCoalescenceLimit)
      {
        #ifdef WARNING
        fprintf(stderr, "nxagentBlockHandler: Flushing with [%d] bytes and count [%d].\n",
                    nxagentFlushPending, nxagentFlushCount - 1);
        #endif

        nxagentFlush = 0;

        nxagentFlushCount = 0;

        nxagentFlushPending = 0;

        NXFlushDisplay(nxagentDisplay, NXFlushLink);
      }
      else
      {
        #ifdef WARNING
        fprintf(stderr, "nxagentBlockHandler: Trying a timeout of [%d] Ms with [%d] bytes and count [%d].\n",
                    nxagentFlushCoalescence / 1000, nxagentFlushPending, nxagentFlushCount);
        #endif

        if (*timeout != NULL)
        {
          #ifdef DEBUG
          fprintf(stderr, "nxagentBlockHandler: Former select timeout was [%ld] Ms.\n",
                      (*timeout) -> tv_sec * 1000 + (*timeout) -> tv_usec / 1000);
          #endif

          (*timeout) -> tv_sec  = 0;
          (*timeout) -> tv_usec = nxagentFlushCoalescence;
        }
        else
        {
          #ifdef DEBUG
          fprintf(stderr, "nxagentBlockHandler: Former select timeout was null.\n");
          #endif

          zero.tv_sec  = 0;
          zero.tv_usec = nxagentFlushCoalescence;

          *timeout = &zero;
        }
      }
    }
  }

  #else /* #ifdef FLUSH_AFTER_MULTIPLE_READS */

  /*
   * We are entering the select. Tell the NX
   * transport to write any produced data to
   * the remote end.
   */

  nxagentFlush = 0;

  NXFlushDisplay(nxagentDisplay, NXFlushLink);

  #endif /* #ifdef FLUSH_AFTER_MULTIPLE_READS */
}

void nxagentWakeupHandler(pointer data, int count, pointer mask)
{
  if (nxagentException.sigHup || nxagentException.ioError)
  {
    #ifdef TEST
    fprintf(stderr,"nxagentWakeupHandler: Got SIGHUP or I/O ERROR.\n");
    #endif

    nxagentHandleConnectionStates();
  }

  if (count > 0)
  {
    /*
     * Try to read more events from the display
     * connection only when WaitForSomething()
     * reported that the display descriptor is
     * ready.
     */

    if (FD_ISSET(nxagentXConnectionNumber, (fd_set *) mask))
    {
      #ifdef TEST
      fprintf(stderr, "nxagentWakeupHandler: Reading the X events with count [%d].\n",
                  count);
      #endif

      nxagentDispatchEvents(NULL);

      count--;
    }
    #ifdef TEST
    else
    {
      fprintf(stderr, "nxagentWakeupHandler: Not reading the X events with count [%d].\n",
                  count);
    }
    #endif
  }

  if (count == 0)
  {
    if (nxagentFlushPending > 0)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentWakeupHandler: Flushing the NX link with [%d] bytes and count [%d].\n",
                  nxagentFlushPending, nxagentFlushCount);
      #endif

      nxagentFlush = 0;

      nxagentFlushCount = 0;

      NXFlushDisplay(nxagentDisplay, NXFlushIdle);
    }
    #ifdef TEST
    else
    {
      fprintf(stderr, "nxagentWakeupHandler: Nothing to do with no pending bytes available.\n");
    }
    #endif
  }
  #ifdef WARNING
  else
  {
    if (nxagentFlushPending > 0)
    {
      fprintf(stderr, "nxagentWakeupHandler: Still reading with [%d] bytes and count [%d].\n",
                  nxagentFlushPending, nxagentFlushCount);
    }
  }
  #endif

  /*
   * If the XDM connection can't be established
   * we'll need to create a dialog to notify the
   * user and give him/her a chance to terminate
   * the session.
   */

  if (nxagentOption(Xdmcp) == 1 && nxagentXdmcpUp == 0)
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentWakeupHandler: XdmcpState [%d].\n", XdmcpState);
    #endif

    if (XdmcpState == XDM_RUN_SESSION)
    {
      nxagentXdmcpUp = 1;
    }

    if (nxagentXdmcpUp == 0)
    {
      #ifdef DEBUG
      fprintf(stderr, "nxagentWakeupHandler: XdmcpTime [%lu].\n",
                  GetTimeInMillis() - XdmcpStartTime);
      #endif

      #ifdef DEBUG
      fprintf(stderr, "nxagentWakeupHandler: XdmcpTimeOutRtx [%d].\n",
                  XdmcpTimeOutRtx);
      #endif

      if (nxagentXdmcpAlertUp == 0 &&
              GetTimeInMillis() - XdmcpStartTime >= XDM_TIMEOUT)
      {
        #ifdef WARNING
        fprintf(stderr, "nxagentWakeupHandler: WARNING! The XDM session seems to be "
                    "unable to start after [%ld] ms.\n", GetTimeInMillis() - XdmcpStartTime);
        #endif

        NXTransAlert(FAILED_XDMCP_CONNECTION_ALERT, NX_ALERT_REMOTE);

        nxagentXdmcpAlertUp = 1;
      }
    }
  }
}
