/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXPROXY, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "ReadBuffer.h"

#include "Transport.h"

//
// Set the verbosity level.
//

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG

ReadBuffer::ReadBuffer(Transport *transport, unsigned int initialReadSize,
                           unsigned int maximumReadSize)

  : transport_(transport), initialReadSize_(initialReadSize),
        maximumReadSize_(maximumReadSize)
{
  size_   = initialReadSize_;
  buffer_ = new unsigned char[size_];

  #ifdef VALGRIND

  memset(buffer_, 0, size_);

  #endif

  length_ = 0;
  start_  = 0;

  remaining_ = 1;
}

ReadBuffer::~ReadBuffer()
{
  delete [] buffer_;
}

void ReadBuffer::partialReset()
{
  if (length_ == 0 && size_ > initialReadSize_)
  {
    #ifdef TEST
    *logofs << "ReadBuffer: Resetting buffer for FD#"
            << transport_ -> fd() << " from " << size_
            << " to " << initialReadSize_ << " bytes.\n"
            << logofs_flush;
    #endif

    delete [] buffer_;

    int newSize = initialReadSize_;

    unsigned char *newBuffer = new unsigned char[newSize];

    if (buffer_ == NULL)
    {
      #ifdef PANIC
      *logofs << "ReadBuffer: PANIC! Can't allocate "
              << newSize << " bytes of memory for buffer "
              << "in context [A].\n" << logofs_flush;
      #endif

      cerr << "Error" << ": Can't allocate memory for "
           << "read buffer in context [A].\n";

      HandleCleanup();
    }

    buffer_ = newBuffer;
    size_   = newSize;

    remaining_ = 1;

    #ifdef VALGRIND

    memset(buffer_, '\0', size_);

    #endif
  }
}

void ReadBuffer::fullReset()
{
  //
  // This is not called at the moment,
  // except when the proxy is reset.
  //

  delete [] buffer_;

  size_   = initialReadSize_;
  buffer_ = new unsigned char[size_];

  if (buffer_ == NULL)
  {
    #ifdef PANIC
    *logofs << "ReadBuffer: PANIC! Can't allocate "
            << size_ << " bytes of memory for buffer "
            << "in context [B].\n" << logofs_flush;
    #endif

    cerr << "Error" << ": Can't allocate memory for "
         << "read buffer in context [B].\n";

    HandleCleanup();
  }

  #ifdef VALGRIND

  memset(buffer_, 0, size_);

  #endif

  length_ = 0;
  start_  = 0;

  remaining_ = 1;
}

int ReadBuffer::readMessage(unsigned int readLength)
{
  //
  // Save the amount of data requested by
  // the caller.
  //

  unsigned int requestedLength = readLength;

  //
  // Determine the best size for the read
  // buffer. As a start try to copy all
  // the data that is available in the
  // transport buffer.
  //

  readLength = transport_ -> pending();

  //
  // If the pending data is not enough to
  // make a complete message, resize the
  // buffer to accomodate the data that
  // is missing.
  //

  if (readLength < remaining_)
  {
    readLength = remaining_;
  }

  //
  // If this is still too few bytes, make
  // space for at least the minimum amount
  // of data for this type of read buffer.
  //

  if (readLength < initialReadSize_)
  {
    readLength = initialReadSize_;
  }

  #ifdef DEBUG
  *logofs << "ReadBuffer: Checking buffer for FD#"
          << transport_ -> fd() << " with requested size "
          << requestedLength << " suggested size " << readLength
          << " buffer size " << size_ << " length " << length_
          << " and start " << start_  << ".\n"
          << logofs_flush;
  #endif
  
  //
  // Be sure that we have enough space to
  // store all the suggested data.
  //

  if (length_ + readLength > size_)
  {
    unsigned int newSize = length_ + readLength;

    #ifdef TEST
    *logofs << "ReadBuffer: Resizing buffer for FD#"
            << transport_ -> fd() << " from " << size_
            << " to " << newSize << " bytes.\n"
            << logofs_flush;
    #endif

    unsigned char *newBuffer = new unsigned char[newSize];

    if (newBuffer == NULL)
    {
      #ifdef PANIC
      *logofs << "ReadBuffer: PANIC! Can't allocate "
              << newSize << " bytes of memory for buffer "
              << "in context [C].\n" << logofs_flush;
      #endif

      cerr << "Error" << ": Can't allocate memory for "
           << "read buffer in context [C].\n";

      HandleCleanup();
    }

    #ifdef VALGRIND

    memset(newBuffer, '\0', newSize);

    #endif

    memcpy(newBuffer, buffer_ + start_, length_);

    delete [] buffer_;

    buffer_ = newBuffer;
    size_   = newSize;
  }
  else if (start_ != 0 && length_ != 0)
  {
    //
    // If any bytes are left due to a partial
    // message, shift them to the beginning
    // of the buffer.
    //
    // TODO: This is a quite inefficient way
    // to make room in the buffer. We know in
    // advance the amount of bytes needed so
    // should always allocate the exact size
    // and only move the small amount of data
    // due to the partial reads.
    //

    #ifdef TEST
    *logofs << "ReadBuffer: Moving " << length_
            << " bytes of data " << "at beginning of "
            << "the buffer for FD#" << transport_ -> fd() 
            << ".\n" << logofs_flush;
    #endif

    memmove(buffer_, buffer_ + start_, length_);
  }

  start_ = 0;

  #ifdef DEBUG
  *logofs << "ReadBuffer: Buffer size is now " << size_ 
          << " length is " << length_ << " and start is " 
          << start_ << ".\n" << logofs_flush;
  #endif

  readLength = size_ - length_;

  if (requestedLength > 0 &&
          readLength > requestedLength)
  {
    readLength = requestedLength;
  }

  if (readLength > maximumReadSize_)
  {
    readLength = maximumReadSize_;
  }

  unsigned char *readData = buffer_ + length_;

  #ifdef DEBUG
  *logofs << "ReadBuffer: Going to read " << readLength 
          << " bytes from FD#" << transport_ -> fd() << ".\n"
          << logofs_flush;
  #endif

  int bytesRead = transport_ -> read(readData, readLength);

  if (bytesRead > 0)
  {
    #ifdef TEST
    *logofs << "ReadBuffer: Read " << bytesRead << " bytes from FD#" 
            << transport_ -> fd() << ".\n" << logofs_flush;
    #endif

    length_ += bytesRead;
  }
  else if (bytesRead < 0)
  {
    #ifdef TEST
    *logofs << "ReadBuffer: Error detected reading from FD#"
            << transport_ -> fd() << ".\n" << logofs_flush;
    #endif

    return -1;
  }
  #ifdef TEST
  else
  {
    *logofs << "ReadBuffer: No data read from FD#"
            << transport_ -> fd() << " with remaining "
            << remaining_ << ".\n" << logofs_flush;
  }
  #endif

  #ifdef TEST
  *logofs << "ReadBuffer: The transport for FD#"
          << transport_ -> fd() << " has " << transport_ ->
             pending() << " bytes pending.\n" << logofs_flush;
  #endif

  return bytesRead;
}

const unsigned char *ReadBuffer::getMessage(unsigned int &controlLength,
                                                unsigned int &dataLength)
{
  if (length_ == 0)
  {
    #ifdef DEBUG
    *logofs << "ReadBuffer: No message can be located for FD#"
            << transport_ -> fd() << ".\n"<< logofs_flush;
    #endif

    return NULL;
  }

  unsigned int trailerLength;

  #ifdef DEBUG
  *logofs << "ReadBuffer: Going to locate message with start at "
          << start_ << " and length " << length_ << " for FD#"
          << transport_ -> fd() << ".\n" << logofs_flush;
  #endif

  int located = locateMessage(buffer_ + start_, buffer_ + start_ + length_,
                                  controlLength, dataLength, trailerLength);

  if (located <= 0)
  {
    //
    // No message or error was detected 
    // uncompressing the message.
    //

    #ifdef DEBUG
    *logofs << "ReadBuffer: No message was located for FD#"
            << transport_ -> fd() << ".\n" << logofs_flush;
    #endif

    return NULL;
  }
  else
  {
    const unsigned char *result = buffer_ + start_;

    if (dataLength > 0)
    {
      //
      // Message contains data, so go to the
      // first byte of payload.
      //

      result += trailerLength;

      start_  += (dataLength + trailerLength);
      length_ -= (dataLength + trailerLength);
    }
    else
    {
      //
      // It is a control message.
      //

      start_  += (controlLength + trailerLength);
      length_ -= (controlLength + trailerLength);
    }

    #ifdef DEBUG
    *logofs << "ReadBuffer: Located message for FD#" << transport_ -> fd()
            << " with control length " << controlLength << " and data length "
            << dataLength << ".\n" << logofs_flush;
    #endif

    remaining_ = 1;

    return result;
  }
}
